-- NOTICE: This file is borrowed from Gediminas Liktaras year 2009 submission.

--------------------------------------------------------------------------------
-- | 
-- Module: Bitmap
--
-- Defines structures to represent a bitmap, as well as basic functions to 
-- manipulate them. Currently the bitmap is implemented as an array of color
-- values.
--
-- Due to the way arrays are implemented in Haskell, introduction of any change
-- into a bitmap will result in a full copy of the array, containing the pixel
-- values. So, setting every pixel individually in a 100x100 bitmap will result 
-- in 10000 copies. 
--
-- To prevent this, I introduced the 'setPixels' function. It modifies multiple
-- pixels at the same time, which copies the pixel array only once. This is why
-- it makes sense to accumulate changes before applying them and why drawing
-- functions in 'BitmapTools' return lists of 'PixelChange'.
--
--------------------------------------------------------------------------------

module Bitmap (
    -- * Types
    Bitmap,
    Pixel,
    PixelChange,
    
    -- * Bitmap creation
    makeBitmap',    -- :: ColorRGB -> Int -> Int -> Bitmap
    makeBitmap,     -- :: Int -> Int -> Bitmap

    -- * Bitmap manipulation
    getBitmapSize,  -- :: Bitmap -> (Int, Int)
    getPixelColor,  -- :: Bitmap -> Pixel -> ColorRGB
    setPixels,      -- :: Bitmap -> [PixelChange] -> Bitmap

    bitmapAsList,   -- :: Bitmap -> [[(Int, Int, Int)]]
    exportAsBMP
    ) where

import Colors

import Data.Array
import Data.Char
import Test.QuickCheck

import Utils
import qualified Data.ByteString.Char8 as B


--------------------------------------------------------------------------------
-- BITMAP TYPE
--------------------------------------------------------------------------------

-- | This type represents a bitmap (i.e. a matrix of colors). 
data Bitmap = BMP (Array (Int,Int) ColorRGB)
              deriving (Eq, Read, Show)

-- | A type, which denotes a location on a 2D plane.
type Pixel = (Int, Int)

-- | A type, which denotes how the given pixel should be modified.
type PixelChange = (Pixel, ColorRGB)


--------------------------------------------------------------------------------
-- BITMAP CREATION
--------------------------------------------------------------------------------

-- | Creates a bitmap with specified dimensions. All pixels are set to the given
-- color.
makeBitmap' :: ColorRGB -> Int -> Int -> Bitmap
makeBitmap' color width height = BMP (array bounds pixels)
    where
    wd = width - 1
    ht = height - 1
    bounds = ((0, 0), (wd, ht))
    pixels = zip [(w,h) | h <- [0..ht], w <- [0..wd]] (cycle [color])

-- | Creates a bitmap with specified dimensions. All pixels are initially black.
makeBitmap :: Int -> Int -> Bitmap
makeBitmap width height = makeBitmap' black width height


--------------------------------------------------------------------------------
-- INFORMATION ABOUT BITMAP
--------------------------------------------------------------------------------

-- | Returns the dimensions of the given bitmap.
getBitmapSize :: Bitmap -> (Int, Int)
getBitmapSize (BMP pxArr) = (w + 1, h + 1)
    where
    ((_,_), (w,h)) = bounds pxArr

-- | Returns the color of the specified pixel.
getPixelColor :: Bitmap -> Pixel -> ColorRGB
getPixelColor (BMP pxArr) px = pxArr ! px

-- | Returns the bitmap as a list of lists of tuples, which represent rows and
-- colors.
bitmapAsList :: Bitmap -> [[(Int, Int, Int)]]
bitmapAsList (BMP pxArr) = [[ color2tuple (pxArr ! (x, y)) | x <- [0..w]] | y <- [0..h] ]
    where
    ((_,_), (w,h)) = bounds pxArr


--------------------------------------------------------------------------------
-- BITMAP MANIPULATION FUNCTIONS
--------------------------------------------------------------------------------

-- | Sets the given pixels to their specified colors. This operation is quite
-- expensive - accumulate changes before introducing them.
setPixels :: Bitmap -> [PixelChange] -> Bitmap
setPixels (BMP pxArr) changes = BMP (pxArr // changes)


-- | Saves the given bitmap as a specified BMP file. 
exportAsBMP :: Bitmap -> String -> IO ()
exportAsBMP bitmap filename = B.writeFile filename (B.pack (header ++ rawData))
    where
    (width, height) = getBitmapSize bitmap
    wd = width - 1
    ht = height - 1

    -- A standard BMP header.
    rowSize  = 4 * ceiling (fromIntegral (24 * width) / 32)
    fileSize = 54 + rowSize * height

    bppField      = reverse (intToRawBin 4 24)
    fileSizeField = reverse (intToRawBin 4 fileSize)
    heightField   = reverse (intToRawBin 4 height)
    rawSizeField  = reverse (intToRawBin 4 (fileSize - 54))
    widthField    = reverse (intToRawBin 4 width)

    header = "\x42\x4D"             -- Magic number (BM).
          ++ fileSizeField          -- Size of the BMP file.
          ++ "\x00\x00"             -- Reserved.
          ++ "\x00\x00"             -- Reserved.
          ++ "\x36\x00\x00\x00"     -- Image data offset (54).
          ++ "\x28\x00\x00\x00"     -- Number of bytes in header (40).
          ++ widthField             -- Width of the bitmap.
          ++ heightField            -- Height of the bitmap.
          ++ "\x01\x00"             -- Number of color planes in use.
          ++ "\x18\x00"             -- Bytes per pixel (24).
          ++ "\x00\x00\x00\x00"     -- Compression method (none).
          ++ rawSizeField           -- Size of raw bitmap data.
          ++ "\x13\x0B\x00\x00"     -- Horizontal resolution of the image (2835).
          ++ "\x13\x0B\x00\x00"     -- Vertical resolution of the image (2835).
          ++ "\x00\x00\x00\x00"     -- Number of colors in the palette.
          ++ "\x00\x00\x00\x00"     -- Important colors (all of them).

    -- Raw image data.
    rowOffsetNum = (4 - (3 * width) `mod` 4) `mod` 4
    rowOffset    = replicate rowOffsetNum '\x00'

    rawData = concat [ concat [rawPixel x y | x <- [0..wd]] ++ rowOffset | y <- reverse [0..ht] ]

    -- Converts the color of a pixel to raw binary (little-endian order).
    rawPixel x y = [chr r, chr g, chr b]
        where
        (RGB r g b) = getPixelColor bitmap (x,y)



