module Main where

import Bitmap
import Colors
import Complex

type BMPFunc = (Int,Int) -> ColorRGB
type BMPFuncGen = (Int,Int) -> BMPFunc

renderFunc :: String -> (Int,Int) -> BMPFuncGen -> IO ()
renderFunc filename size func = exportAsBMP bmp filename
  where
    (w,h) = size
    cfunc = func (w,h)
    chset = [((x,y), cfunc (x,y)) | x <- [0..w-1], y <- [0..h-1]]
    bmp = setPixels (makeBitmap w h) chset

mkRGB x y z = makeRGB (floor (z * 255)) (floor (y * 255)) (floor (x * 255))

complexPlotFunc :: (Complex Double -> Complex Double) -> BMPFuncGen
complexPlotFunc mf (w,h) = f
  where
    x1 = 3.5
    x2 = -2.0
    y1 = -2.5
    y2 = 3.0
    f (x,y) = complexToStyledColor (mf ((x1 + (x2 - x1) * (fromIntegral x / fromIntegral w)) :+
      (y2 + (y1 - y2) * (fromIntegral y / fromIntegral h))))

complexToStyledColor :: Complex Double -> ColorRGB
complexToStyledColor q = color
  where
    piLoop x | x < 0 = piLoop (x + 2*pi)
             | otherwise = x
    mgn = magnitude q
    angle = piLoop (phase q - pi/3)
    sat = (sin (2 * log mgn) + 1) / 2
    color = hueToColor (angle / (2 * pi)) (0.2 + 0.6 * sat) 1

hueToColor :: Double -> Double -> Double -> ColorRGB
hueToColor hue sat val
  | a < 0 = h 0 0 0
  | a < 1 = h 1 a 0
  | a < 2 = h (2 - a) 1 0
  | a < 3 = h 0 1 (a - 2)
  | a < 4 = h 0 (4 - a) 1
  | a < 5 = h (a - 4) 0 1
  | a < 6 = h 1 0 (6 - a)
  | otherwise = h 0 0 0
  where
    a = 6*hue
    c = sat * val
    m = val - c
    h r g b = mkRGB (c * r + m) (c * g + m) (c * b + m)

niceFunction :: Complex Double -> Complex Double
niceFunction x = (x ^ 2 - 1) * (x - (2 :+ 1)) ^ 2 / (x ^ 2 + (2 :+ 2))

main = do putStrLn "Generating image..."
          renderFunc "output.bmp" (600,600) $ complexPlotFunc $ niceFunction

-- Debugging stuff

testFunc :: (Int,Int) -> (Int,Int) -> ColorRGB
testFunc (w,h) = f
  where
    f :: (Int,Int) -> ColorRGB
    f (x,y) = mkRGB (fromIntegral x / fromIntegral w) (fromIntegral y / fromIntegral h) 0

testGrayscaleBar (w,h) (x,y) = mkRGB coef coef coef
  where coef = (fromIntegral x / fromIntegral w)
