-- NOTICE: This file is borrowed from Gediminas Liktaras year 2009 submission.

--------------------------------------------------------------------------------
-- | 
-- Module: Utils
--
-- Defines various miscellaneous functions and structures.
-- 
--------------------------------------------------------------------------------

module Utils (
    -- * Constants
    epsilon,        -- :: Double

    -- * Rectangle functions
    Rectangle(..),
    makeRect,       -- :: (Real a) => (a, a) -> (a, a) -> Rectangle a
    pointInRect,    -- :: (Real a) => (a, a) -> Rectangle a -> Bool

    -- * Conversions to/from raw binary/hexadecimal strings
    intToHex,       -- :: Int -> Int -> String
    intToRawBin,    -- :: Int -> Int -> String
    rawBinToInt,    -- :: String -> Int

    -- * Miscellaneous functions
    minMax,     -- :: (Ord a) => a -> a -> (a, a)
    partition,  -- :: [a] -> Int -> [[a]]
    restrict,   -- :: (Ord a) => a -> a -> a -> a
    slice       -- :: [a] -> Int -> Int -> [a]

    ) where

import Data.Char (chr, intToDigit, ord, toUpper)


--------------------------------------------------------------------------------
-- RECTANGLE
--------------------------------------------------------------------------------

-- | Accuracy of floating-point operations.
epsilon :: Double
epsilon = 1e-6


--------------------------------------------------------------------------------
-- RECTANGLE
--------------------------------------------------------------------------------

-- | Rectangle type.
data (Real a) => Rectangle a = Rect (a, a) (a, a) 
                              deriving(Eq, Read, Show)

-- | Creates a rectangle object from the coordinates of any two of its opposite
-- corners.
makeRect :: (Real a) => (a, a) -> (a, a) -> Rectangle a
makeRect (x1, y1) (x2, y2) = Rect (p1x, p1y) (p2x, p2y)
    where
    (p1x, p2x) = minMax x1 x2
    (p1y, p2y) = minMax y1 y2

-- | Returns True, if the specified point is within the rectangle.
pointInRect :: (Real a) => (a, a) -> Rectangle a -> Bool
pointInRect (x, y) (Rect (ulx, uly) (lrx, lry)) =
    ((x >= ulx) && (x <= lrx) && (y >= uly) && (y <= lry))


--------------------------------------------------------------------------------
-- INTEGER CONVERSION TO/FROM STRINGS
--------------------------------------------------------------------------------

-- | Returns the given integer as a hexadecimal field of specified size, aligned 
-- to the left. 
intToHex :: Int -> Int -> String
intToHex fieldSize val = if hexStrLen >= fieldSize
                         then hexStr
                         else (replicate (fieldSize - hexStrLen) '0') ++ hexStr
    where
    hexStr    = toHex val
    hexStrLen = length hexStr

    toHex 0 = ""
    toHex x = toHex restDig ++ [(toUpper . intToDigit) cDig]
        where
        cDig    = x `mod` 16
        restDig = x `div` 16


-- | Returns the given integer as a raw binary field of specified size.
-- Big-endian order is used to order bytes - use function reverse, to obtain the
-- Little-endian order.
intToRawBin :: Int -> Int -> String
intToRawBin fieldSize val = if rawStrLen >= fieldSize
                            then rawStr
                            else (replicate (fieldSize - rawStrLen) '\x00') ++ rawStr
    where
    rawStr    = toRawBin val
    rawStrLen = length rawStr

    toRawBin 0 = ""
    toRawBin x = toRawBin restDig ++ [chr cDig]
        where
        cDig    = x `mod` 256
        restDig = x `div` 256


-- | Converts a binary string into a corresponding integer.
rawBinToInt :: String -> Int
rawBinToInt xs
    | length xs > 4 = error "Cannot convert such a large value."
    | otherwise     = fromRawBin (reverse xs)
    where
    fromRawBin []     = 0
    fromRawBin (x:xs) = (ord x) + 256 * (fromRawBin xs)


--------------------------------------------------------------------------------
-- MISCELLANEOUS FUNCTIONS
--------------------------------------------------------------------------------

-- | Takes two values and returns them as a pair. The first element of the pair
-- is smaller of the two values, the second element is the larger one.
minMax :: (Ord a) => a -> a -> (a, a)
minMax x y = if (x > y) then (y, x) else (x, y)

-- | Splits a list into sub-lists of specified size.
partition :: [a] -> Int -> [[a]]
partition xs n
    | length xs <= n = [xs]
    | otherwise      = (take n xs) : (partition (drop n xs) n)

-- | Makes sure that the given variable is within the specified interval.
restrict :: (Ord a) => a -> a -> a -> a
restrict lBound uBound x
    | x < lBound = lBound
    | x > uBound = uBound
    | otherwise  = x

-- | Returns a slice of the list.
slice :: [a] -> Int -> Int -> [a]
slice xs from to = (drop from . take (to+1)) xs

